from __future__ import absolute_import
from __future__ import unicode_literals

from blazeutils.testing import raises

from keg.app import Keg, KegAppError


class TestInit(object):

    @raises(KegAppError, 'init() already called')
    def test_init_called_twice_error(self):
        app = Keg(__name__)
        app.init()
        app.init()



from __future__ import absolute_import
from __future__ import unicode_literals

import mock
import pytest

from keg_apps.cli import CLIApp
from keg_apps.db.app import DBApp
from keg.testing import CLIBase


class TestCLI(CLIBase):
    app_cls = CLIApp
    cmd_name = 'hello'

    def test_class_command(self):
        result = self.invoke()
        assert 'hello keg test' in result.output

    def test_arg_command(self):
        result = self.invoke(cmd_name='foo2')
        assert 'hello foo' in result.output

    def test_argument_passing(self):
        result = self.invoke('bar', cmd_name='foo2')
        assert 'hello bar' in result.output

    def test_missing_command(self):
        result = self.invoke(cmd_name='baz', exit_code=2)
        assert 'No such command "baz"' in result.output

    @pytest.mark.skipif(True, reason='reminder')
    def test_default_exception_handling(self):
        """
            invoke_command() does some stuff to help with exceptions, it should be tested or
            removed.  So comment in that function for details.
        """


class TestConfigCommand(CLIBase):
    app_cls = CLIApp
    cmd_name = 'develop config'

    def test_output(self):
        result = self.invoke()
        # use a value that should be towards the end of the output as a reasonble indicator the
        # command worked as expected.
        assert 'USE_X_SENDFILE = False' in result.output

    def test_database_disabled(self):
        result = self.invoke(cmd_name='develop db')
        assert 'Database not enabled for this app.  No subcommands available.' in result.output
        assert 'Lists database related sub-commands.' not in result.output


class TestDatabaseCommands(CLIBase):
    app_cls = DBApp
    cmd_name = 'develop db'

    def test_help_info(self):
        result = self.invoke()
        assert 'Lists database related sub-commands.' in result.output
        assert 'init   Create all db objects, send related events.' in result.output

    @mock.patch('keg.db.DatabaseManager.db_init')
    def test_init(self, m_db_init):
        result = self.invoke('init')
        assert 'Database initialzed' in result.output
        m_db_init.assert_called_once_with()

    @mock.patch('keg.db.DatabaseManager.db_init_with_clear')
    def test_init_with_clear(self, m_db_init_wc):
        result = self.invoke('init', '--clear-first')
        assert 'Database cleared and initialized' in result.output
        m_db_init_wc.assert_called_once_with()

    @mock.patch('keg.db.DatabaseManager.db_clear')
    def test_clear(self, m_db_clear):
        result = self.invoke('clear')
        assert 'Database cleared' in result.output
        m_db_clear.assert_called_once_with()



from __future__ import absolute_import
from __future__ import unicode_literals

import logging
from logging.handlers import RotatingFileHandler
import os.path as osp

from keg.utils import ensure_dirs


class Logging(object):

    def __init__(self, config):
        self.config = config
        self.log_dpath = self.config['KEG_LOG_DPATH']
        self.log_fname = self.config['KEG_LOG_FNAME']
        self.file_format_str = self.config['KEG_LOG_FILE_FORMAT_STR']
        self.stdout_format_str = self.config['KEG_LOG_STDOUT_FORMAT_STR']

        self.loggers = [logging.getLogger(location) for location in self.logged_locations]
        self.app_logger = logging.getLogger(self.config.app_import_name)

    @property
    def logged_locations(self):
        # todo: could probably make this configurable
        return [
            self.config.app_import_name,
            'keg',
        ]

    def init_logger_level(self):
        for logger in self.loggers:
            # Our stdout logger uses INFO, so make that the level we log at.  The handler levels
            # on the file logger will keep INFO messages out of the log file.
            # If we ever impliment a debug or verbose flag on the command line, we'll have to adjust
            # this accordingly.
            logger.setLevel(logging.INFO)

    def log_fpath(self):
        return osp.join(self.log_dpath, self.log_fname)

    def init_main_log(self):
        log_fpath = self.log_fpath()
        file_formatter = logging.Formatter(self.file_format_str)

        dir_mode = self.config['KEG_DIR_MODE']
        ensure_dirs(self.log_dpath, mode=dir_mode)

        log_max = self.config['KEG_LOG_MAX_BYTES']
        log_backups = self.config['KEG_LOG_MAX_BACKUPS']

        file_handler = RotatingFileHandler(log_fpath, maxBytes=log_max, backupCount=log_backups,
                                           encoding='utf-8')
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(logging.WARN)

        for logger in self.loggers:
            logger.addHandler(file_handler)

    def init_stdout_log(self):
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(self.stdout_format_str))
        handler.setLevel(logging.INFO)

        for logger in self.loggers:
            logger.addHandler(handler)

    def init_app(self):
        self.init_logger_level()
        self.init_main_log()
        self.init_stdout_log()



from __future__ import absolute_import
from __future__ import unicode_literals

import flask

keg = flask.Blueprint('keg', __name__)


@keg.route('/exception-test')
def exception_test():
    raise Exception('Deliberate exception for testing purposes')


@keg.route('/ping')
def ping():
    return '{} ok'.format(flask.current_app.name)



from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from keg.app import Keg

log = logging.getLogger(__name__)


class LoggingApp(Keg):
    import_name = __name__



from __future__ import absolute_import
from __future__ import unicode_literals

from keg.app import Keg


class DBApp(Keg):
    import_name = 'keg_apps.db'
    db_enabled = True


if __name__ == '__main__':
    DBApp.cli_run()



from __future__ import absolute_import
from __future__ import unicode_literals

import getpass
import re
import sys

import six

try:
    import keyring
except ImportError:
    keyring = None


class KeyringError(Exception):
    pass


class Manager(object):

    # regex that matches "${*}$" where * = any printable ascii character
    # that isn't "}"
    # see: http://www.catonmat.net/blog/my-favorite-regex/
    sub_pattern = '(\$\{([ -|~]+?)\}\$)'

    backend_min_priority = 1

    def __init__(self, app):
        self.sub_re = re.compile(self.sub_pattern)
        self.log = app.logger
        self.app = app
        self.sub_keys_seen = set()

        if not self.verify_backend():
            self.log.warning(
                'Insecure keyring backend detected, keyring substitution unavailable. '
                'Run this app\'s keyring command for more info.'
            )

    def verify_backend(self):
        backend = keyring.get_keyring()
        if backend.priority < self.backend_min_priority:
            return False
        return True

    def substitute(self, data):
        if not self.verify_backend():
            return

        for key in data.keys():
            value = data[key]
            if not isinstance(value, six.string_types):
                continue
            matches = self.sub_re.finditer(value)
            for match in matches:
                replace_this = match.group(1)
                keyring_key = match.group(2)

                self.sub_keys_seen.add(keyring_key)

                self.log.debug('keyring substitute: replacing {0} for data key "{1}"'
                               .format(replace_this, key))

                # reassign value for cases when there is more than one keyring replacement
                # needed in a config value
                value = data[key]

                keyring_value = keyring.get_password(self.app.name, keyring_key)
                if keyring_value is None:
                    msg = 'Enter value for "{0}": '.format(keyring_key)
                    # Windows getpass implementation breaks in Python2 when passing it unicode
                    if sys.version_info < (3,) and sys.platform.startswith('win'):
                        msg = str(msg)
                    keyring_value = getpass.getpass(msg)
                    keyring.set_password(self.app.name, keyring_key, keyring_value)
                data[key] = value.replace(replace_this, keyring_value, 1)

    def delete(self, key):
        keyring.delete_password(self.app.name, key)



from __future__ import absolute_import
from __future__ import unicode_literals

from keg.app import Keg
from keg_apps.web.views import all_blueprints


def addkeg(value):
    return '{}Keg'.format(value)


def sayfoo():
    return 'foo'


class WebApp(Keg):
    import_name = __name__
    use_blueprints = all_blueprints
    keyring_enabled = False

    template_filters = {'addkeg': addkeg}
    template_globals = {'sayfoo': sayfoo}



from __future__ import absolute_import
from __future__ import unicode_literals

from flask.signals import Namespace

_signals = Namespace()

# Call when application initializations is completed.
# Sender will be the application instance.
app_ready = _signals.signal('app-ready')

config_ready = _signals.signal('config-ready')
testing_run_start = _signals.signal('testing-run-start')

db_clear_pre = _signals.signal('db-clear-pre')
db_clear_post = _signals.signal('db-clear-post')
db_init_pre = _signals.signal('db-init-pre')
db_init_post = _signals.signal('db-init-post')



from __future__ import absolute_import
from __future__ import unicode_literals

from collections import defaultdict
import platform

import click
import flask
from six.moves import urllib

from keg import current_app
from keg._flask_cli import FlaskGroup, script_info_option, with_appcontext, run_command, \
    shell_command
from keg.keyring import keyring as keg_keyring


class KegGroup(FlaskGroup):
    def __init__(self, add_default_commands=True, *args, **kwargs):
        FlaskGroup.__init__(self, add_default_commands=False, *args, **kwargs)
        if add_default_commands:
            self.add_command(dev_command)


@click.group('develop', help='Developer info and utils.')
def dev_command():
    pass

dev_command.add_command(run_command)
dev_command.add_command(shell_command)


@dev_command.command('routes', short_help='List the routes defined for this app.')
@with_appcontext
def routes_command():
    output = []
    endpoint_len = 0
    methods_len = 0

    # calculate how wide the output columns need to be
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        endpoint_len = max(endpoint_len, len(rule.endpoint))
        methods_len = max(methods_len, len(methods))

    # generate the output
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        line = urllib.parse.unquote("{}   {}   {}".format(
            rule.endpoint.ljust(endpoint_len), methods.ljust(methods_len), rule))
        output.append(line)

    for line in sorted(output):
        click.echo(line)


@dev_command.command('templates', short_help='Show paths searched for a template.')
@with_appcontext
def templates_command():
    jinja_loader = flask.current_app.jinja_env.loader
    paths = defaultdict(list)
    for template in jinja_loader.list_templates():
        for loader, template in jinja_loader._iter_loaders(template):
            for dpath in loader.searchpath:
                paths[dpath].append(template)
    for dpath in sorted(paths.keys()):
        template_paths = sorted(paths[dpath])
        click.echo(dpath)
        for tpath in template_paths:
            click.echo('    {}'.format(tpath))


@dev_command.command('config', short_help='List info related to config files, profiles, and'
                     ' values.')
@with_appcontext
def config_command():
    app = flask.current_app
    config = app.config
    keys = list(config.keys())
    keys.sort()

    click.echo('Default config objects:')
    for val in config.default_config_locations_parsed():
        click.echo('    {}'.format(val))

    click.echo('Config file locations:')
    for val in config.config_file_paths():
        click.echo('    {}'.format(val))

    click.echo('Config objects used:')
    for val in config.configs_found:
        click.echo('    {}'.format(val))

    click.echo('Resulting app config (including Flask defaults):')
    for key in keys:
        click.echo('    {} = {}'.format(key, config[key]))


class DatabaseGroup(click.MultiCommand):

    def list_commands(self, ctx):
        return ['clear', 'init']

    def get_command(self, ctx, name):
        if name == 'init':
            return database_init
        if name == 'clear':
            return database_clear


@dev_command.command('db', cls=DatabaseGroup, invoke_without_command=True,
                     help='Lists database related sub-commands.')
@with_appcontext
@click.pass_context
def database_group(ctx):
    # only take action if no subcommand is involved.
    if ctx.invoked_subcommand is None:
        if not current_app.db_enabled:
            click.echo('Database not enabled for this app.  No subcommands available.')
        else:
            # Database enabled, but no subcommand was given, therefore we want to just show
            # the help message, which would be the default behavior if we had not used the
            # invoke_Without_command option.
            click.echo(ctx.get_help())
            ctx.exit()


@click.command('init', short_help='Create all db objects, send related events.')
@click.option('--clear-first', default=False, is_flag=True,
              help='Clear DB of all data and drop all objects before init.')
@with_appcontext
def database_init(clear_first):
    if clear_first:
        current_app.db_manager.db_init_with_clear()
        click.echo('Database cleared and initialized')
    else:
        current_app.db_manager.db_init()
        click.echo('Database initialzed')


@click.command('clear', short_help='Clear DB of all data and drop all objects.')
@with_appcontext
def database_clear():
    current_app.db_manager.db_clear()
    click.echo('Database cleared')


class KeyringGroup(click.MultiCommand):

    def list_commands(self, ctx):
        if keg_keyring:
            return ['delete', 'list-keys', 'setup', 'status']
        else:
            return ['status']

    def get_command(self, ctx, name):
        if name == 'status':
            return keyring_status
        if name == 'list-keys':
            return keyring_list_keys
        if name == 'delete':
            return keyring_delete


def keyring_notify_no_module():
    click.echo('Keyring module not installed. Keyring functionality disabled.\n\nYou can'
               ' enable keyring functionality by installing the package:'
               ' `pip install keyring`.')


@dev_command.command('keyring', cls=KeyringGroup, invoke_without_command=True,
                     help='Lists keyring related sub-commands.')
@click.pass_context
def keyring_group(ctx):
    # only take action if no subcommand is involved.
    if ctx.invoked_subcommand is None:
        if keg_keyring is None:
            keyring_notify_no_module()
        else:
            # keyring is available, but no subcommand was given, therefore we want to just show
            # the help message, which would be the default behavior if we had not used the
            # invoke_Without_command option.
            click.echo(ctx.get_help())
            ctx.exit()


@click.command('status', short_help='Show keyring related status info.')
@click.option('--unavailable', default=False, is_flag=True,
              help='Show unavailable backends with reasons.')
@with_appcontext
def keyring_status(unavailable):
    if keg_keyring is None:
        keyring_notify_no_module()
        return
    import keyring
    import keyring.backend as kb
    viable = kb.get_all_keyring()

    # call get_all_keyring() before this so we are sure all keyrings are loaded
    # on KeyringBackend
    if unavailable:
        click.echo('Unavailable backends')
        for cls in kb.KeyringBackend._classes:
            try:
                cls.priority
            except Exception as e:
                click.echo('    {0.__module__}:{0.__name__} - {1}'.format(cls, e))

    click.echo('\nAvailable backends (backends with priority < 1 are not'
               ' recommended and may be insecure)')
    for backend in viable:
        click.echo('    {0.__module__}:{0.__name__} (priority: {1})'
                   .format(backend.__class__, backend.priority))

    click.echo('\nDefault backend')
    backend = keyring.get_keyring()
    click.echo('    {0.__module__}:{0.__name__}'.format(backend.__class__))
    if hasattr(backend, 'file_path'):
        click.echo('    file path: {}'.format(backend.file_path))

    if not flask.current_app.keyring_enabled:
        click.echo('\nKeyring functionality for this app has been DISABLED through the config'
                   ' setting KEG_KEYRING_ENABLE.')
    elif not flask.current_app.keyring_manager.verify_backend():
        click.echo('\nWARNING: the current backend is insecure,'
                   ' keyring substitution unavailable.')
        if platform.system() == 'Linux':
            click.echo('\nTRY THIS: use the SecretStorage Setup utility to get a more secure'
                       ' keyring backend.')
            click.echo('https://pypi.python.org/pypi/SecretStorage-Setup\n')


@click.command('list-keys', short_help='Show all keys used in config value substitution.')
@with_appcontext
def keyring_list_keys():
    km = flask.current_app.keyring_manager
    for key in sorted(km.sub_keys_seen):
        click.echo(key)


@click.command('delete', short_help='Delete an entry from the keyring.')
@click.argument('key')
@with_appcontext
def keyring_delete(key):
    flask.current_app.keyring_manager.delete(key)


def init_app_cli(appcls):

    # this function will be used to initialize the app along, including the value of config_profile
    # which can be passed on the command line.
    def _create_app(script_info):
        app = appcls()
        app.init(config_profile=script_info.data['config_profile'])
        return app

    @click.group(cls=KegGroup, create_app=_create_app)
    @script_info_option('--profile', script_info_key='config_profile', default=None,
                        help='Name of the configuration profile to use.')
    def cli(**kwargs):
        pass

    return cli



from __future__ import absolute_import
from __future__ import unicode_literals

import pathlib

from keg_apps.logging import LoggingApp, log
from keg import signals


class TestLogging(object):

    def test_default_log_file(self, tmpdir):
        tmpdir = str(tmpdir)

        @signals.config_ready.connect
        def apply_config(app):
            app.config['KEG_LOG_DPATH'] = tmpdir

        app = LoggingApp().init()
        log.warning('test warn log')
        log.info('test info log')

        log_fpath = pathlib.Path(app.logging.log_fpath())
        with log_fpath.open('r', encoding='utf-8') as fh:
            contents = fh.read()

        assert 'test warn log' in contents
        assert 'test info log' not in contents




class DefaultProfile(object):
    KEG_KEYRING_ENABLE = False


class TestProfile(object):
    KEG_DB_DIALECT_OPTIONS = {
        'postgresql.schemas': ('public', 'fooschema'),
    }

    # These binds are used to test keg dialect functionality. All other database tests
    # are done on the primary connection, which defaults to SQLite.
    SQLALCHEMY_BINDS = {
        # The postgres connection defaults to what our travis build needs.  If testing as a
        # developer, then you should setup a TestProfile config for these tests in
        # file: ~/.config/keg_apps.db/keg_apps.db-config.py
        'postgres': 'postgresql://postgres@localhost/travis_ci_test',
        'sqlite2': 'sqlite:///'
    }



from __future__ import absolute_import
from __future__ import unicode_literals

import six

from blazeutils.strings import simplify


def object_from_source(fpath, objname):
    if six.PY2:
        import imp
        module_name = 'keg_{}'.format(simplify(fpath))
        tmpmodule = imp.load_source(module_name, fpath)
        if not hasattr(tmpmodule, objname):
            return None
        return getattr(tmpmodule, objname)


def with_metaclass(meta, *bases):
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.  Because of internal type checks
    # we also need to make sure that we downgrade the custom metaclass
    # for one level to something closer to type (that's why __call__ and
    # __init__ comes back from type etc.).
    #
    # This has the advantage over six.with_metaclass in that it does not
    # introduce dummy classes into the final MRO.
    #
    # copied from Flask._compat
    class _MetaClass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return _MetaClass(str('temporary_class'), None, {})
